## Data structures for manipulating formulae.
from lit import Lit

class BoolExp:
  def __init__(self, tbl, val):
    self.tbl = tbl
    self.val = val

  def kind(self):
    return self.tbl.get_kind(self.val)

  def children(self):
    childer = self.tbl.get_children(self.val)
    if self.tbl.get_kind(self.val) == ExpTable.LEAF:
      return childer
    else:
      return [ BoolExp(self.tbl, c) for c in self.tbl.get_children(self.val) ]

  def __and__(self, other):
    assert self.tbl == other.tbl
    return BoolExp(self.tbl, self.tbl.apply(ExpTable.AND, self.val, other.val))

  def __or__(self, other):
    assert self.tbl == other.tbl
    return BoolExp(self.tbl, self.tbl.apply(ExpTable.OR, self.val, other.val))

  def __invert__(self):
    return BoolExp(self.tbl, self.tbl.apply(ExpTable.NOT, self.val))

  def __str__(self):
    return self.tbl.exp_str(self.val)

class ExpTable:
  (LEAF, AND, OR, NOT) = (0, 1, 2, 3)
  (T, F) = (1, -1)

  def __init__(self):
    self.cache = {}
    self.table = []

  def ttt(self):
    return BoolExp(self, self.apply(ExpTable.LEAF, ExpTable.T))
  def fff(self):
    return BoolExp(self, self.apply(ExpTable.LEAF, ExpTable.F))

  def leaf(self, lit):
    return BoolExp(self, self.apply(ExpTable.LEAF, lit.v))

  def get_kind(self, val):
    return self.table[val][0]

  def get_children(self, val):
    return self.table[val][1]

  def apply(self, op, c0, c1 = None):
    if (op, c0, c1) in self.cache:
      return self.cache[(op, c0, c1)]
    else:
      val = len(self.table)
      children = [c0]
      if c1 is not None:
        children.append(c1)
      self.table.append( (op, children) )
      self.cache[(op, c0, c1)] = val
      return val

  def exp_str(self, val):
    (op, cs) = self.table[val] 
    if op == ExpTable.LEAF:
      if cs[0] == ExpTable.T:
        return "T"
      elif cs[0] == ExpTable.F:
        return "F"
      else:
        return "x{0}".format(cs[0])
    elif op == ExpTable.NOT:
      return "~" + self.exp_str(cs[0])
    elif op == ExpTable.AND:
      return "".join(["(", self.exp_str(cs[0]), " & ", self.exp_str(cs[1]), ")"])
    elif op == ExpTable.OR:
      return "".join(["(", self.exp_str(cs[0]), " | ", self.exp_str(cs[1]), ")"])
