#!/usr/bin/python

## Code for converting a circuit into a SAT instance
## which produces an n x m lattice, by unfolding the
## quantified variables.
import sys
import qbf
from constraints import sorter, cardnet, atmost_one, reify #, lex_leq
from formula import ExpTable, BoolExp
from itertools import chain
from espresso import dual_prods
import time

#start_time_prep1=time.time()
target_idx = 0

ncols = int(sys.argv[-1])
nrows = int(sys.argv[-2])

problem = qbf.QBF()

symbreak = False
#if "-sym" in sys.argv[1:-2]:
#  symbreak = True
#symbreak = True

## Assume input is in pla format.
invars = set()

prods = []
for line in sys.stdin.readlines():
  [sin, sout] = line.strip().split()
  if sout[target_idx] == '1':
    prods.append(sin)
    for ti in xrange(len(sin)):
      vi = ti+1
      if sin[ti] == '0' or sin[ti] == '1':
        invars |= {vi}

## Convert the set of input variables
## to a list.
invars = sorted(list(invars))

## Trim the products to the care-set
prods = map(lambda prod: "".join([prod[vi-1] for vi in invars]), prods)
duals = dual_prods(prods)

## Set up the expression table
table = ExpTable()
def toExp(v):
  return table.leaf(v)

## Construct the quantified variables
## Ordering [F, T, x0, ~x0, x1, ~x1, ...]
latvars = { (i, j) : problem.existFD(2 + 2*len(invars))
              for i in xrange(nrows) for j in xrange(ncols) }
## qvars = { v : problem.univ() for v in invars }

## Symmetry breaking
#if 1:
#  left_vars = [ latvars[(i, j)] for i in xrange(nrows) for j in xrange(ncols/2) ]
#  right_vars = [ latvars[(i, ncols-j-1)] for i in xrange(nrows) for j in xrange(ncols/2) ]
#  print >> sys.stderr,"left : ", left_vars
#  print >> sys.stderr, "right : ",right_vars
  #lex_leq(problem, left_vars, right_vars)
  #print >> sys.stderr, "left new : ",left_vars
  #print >> sys.stderr, "right new : ",right_vars
#  top_vars = [ latvars[(i, j)] for j in xrange(ncols) for i in xrange(nrows/2) ]
#  bot_vars = [ latvars[(nrows-i-1, j)] for j in xrange(ncols) for i in xrange(nrows/2) ]
#  print >> sys.stderr, "top : ",top_vars
#  print >> sys.stderr, "bot : ",bot_vars
  #lex_leq(problem, top_vars, bot_vars)
  #print >> sys.stderr, "top new : ", top_vars
  #print >> sys.stderr, "bot new : ", bot_vars

#for (i, j) in latvars:
#  print >> sys.stderr, (i, j), ":", latvars[(i, j)]

## Collect the path variables adjacent to a square.
def pos_adjvars(vs, i, j):
  if i == 0:
    return [ vs[(i+1, j)] ]
  if i == nrows-1:
    return [ vs[(i-1, j)] ]
  return [ vs[elt] for elt in [(i-1, j), (i+1, j), (i, j-1), (i, j+1)] if elt in vs ]

def neg_adjvars(vs, i, j):
  if j == 0:
    return [ vs[(ip, j+1)] for ip in [i-1, i, i+1] if (ip, j+1) in vs ]
  if j == ncols-1:
    return [ vs[(ip, j-1)] for ip in [i-1, i, i+1] if (ip, j-1) in vs ]
  return [ vs[elt] for elt in 
            [(i-1, j-1), (i-1, j), (i-1, j+1), (i, j-1), (i, j+1), (i+1, j-1), (i+1, j), (i+1,j+1)]
            if elt in vs ]

## Build the positive instances for a given assignment
def build_pos(prod):
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    ## ~switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    assert len(lvar) == 2*len(prod) + 2

#    """
    problem.addClause([~lvar[0], ~nv])
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        problem.addClause([~lvar[2*vi + 2], ~nv])
      elif prod[vi] == '1':
        problem.addClause([~lvar[2*vi + 3], ~nv])
      else:
        # Don't-cares can't be used
        problem.addClause([~lvar[2*vi + 2], ~nv])
        problem.addClause([~lvar[2*vi + 3], ~nv])

    """
    sv = problem.exist()
    problem.addClause([~lvar[0], ~sv])
    problem.addClause([~lvar[1], sv])
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        problem.addClause([~lvar[2*vi + 2], ~sv])
        problem.addClause([~lvar[2*vi + 3], sv])
      elif prod[vi] == '1':
        problem.addClause([~lvar[2*vi + 3], ~sv])
        problem.addClause([~lvar[2*vi + 2], sv])
      else:
        # Don't-cares can't be used
        problem.addClause([~lvar[2*vi + 2], ~sv])
        problem.addClause([~lvar[2*vi + 3], ~sv])
#    problem.addClause([~sv, lvar[1]]
#        + [ lvar[2*vi + 2] for vi in xrange(len(prod)) if prod[vi] == '1' ]
#        + [ lvar[2*vi + 3] for vi in xrange(len(prod)) if prod[vi] == '0' ])

    problem.addClause([sv, ~nv])
#  """

  if nrows == 1:
    problem.addClause(nvars.values())
  elif ncols == 1:
    for val in nvars.values():
      problem.addClause([val])
  else:
    for (i, j) in nvars.keys():
      nv = nvars[(i, j)]
      adjs = pos_adjvars(nvars, i, j)
      assert len(adjs) in [1, 3, 4]
      
      if len(adjs) == 1:
        ## Terminal edge.
        problem.addClause([adjs[0],~nv])
      else:
        assert len(adjs) >= 3
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        ## nv -> sum(adjs) = 2
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])

    ## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == nrows-1 ], problem.ttt())
    
## And the negative paths
def build_neg(prod):
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    ## switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    assert len(lvar) == 2*len(prod) + 2

#    """
    problem.addClause([~lvar[1], ~nv]) ## [[ lij = T ]] => ~neg[i,j]
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        problem.addClause([~lvar[2*vi + 3], ~nv])
      elif prod[vi] == '1':
        problem.addClause([~lvar[2*vi + 2], ~nv])
      else:
        # Don't care for this product
        problem.addClause([~lvar[2*vi + 3], ~nv])
        problem.addClause([~lvar[2*vi + 2], ~nv])
    """
    sv = problem.exist()
    problem.addClause([~lvar[0], sv])
    problem.addClause([~lvar[1], ~sv])
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        problem.addClause([~lvar[2*vi + 3], ~sv])
        problem.addClause([~lvar[2*vi + 2], sv])
      elif prod[vi] == '1':
        problem.addClause([~lvar[2*vi + 2], ~sv])
        problem.addClause([~lvar[2*vi + 3], sv])
      else:
        # Don't-cares can't be used
        problem.addClause([~lvar[2*vi + 3], ~sv])
        problem.addClause([~lvar[2*vi + 2], ~sv])
#    problem.addClause([~sv, lvar[0]]
#        + [ lvar[2*vi + 3] for vi in xrange(len(prod)) if prod[vi] == '1' ]
#        + [ lvar[2*vi + 2] for vi in xrange(len(prod)) if prod[vi] == '0' ])

    problem.addClause([sv, ~nv])
#    """


  if ncols == 1:
    problem.addClause(nvars.values())
  elif nrows == 1:
    for val in nvars.values():
      problem.addClause([val])
  else:
    for (i, j) in nvars.keys():
      ## Connect up the path variables.
      nv = nvars[(i, j)]
      adjs = neg_adjvars(nvars, i, j)
      assert (len(adjs) in [2, 3, 5, 8])
      
      if len(adjs) <= 2:
        ## Terminal edge.
        assert j in [0, ncols-1]
        #adjv = problem.exist()
        #atmost_one(problem, adjs, adjv)
        #problem.addClause([adjv, ~nv])
        problem.addClause([adjs[0],adjs[1], ~nv])
        problem.addClause([~adjs[0],~adjs[1], ~nv])
        problem.addClause([adjs[0],adjs[1], ~nv])
      elif len(adjs) == 3:
        ## Terminal edge.
        assert j in [0, ncols-1]
        #adjv = problem.exist()
        #atmost_one(problem, adjs, adjv)
        #problem.addClause([adjv, ~nv])
        
        #problem.addClause([adjs[0],adjs[1],adjs[2], ~nv])
        adjv = problem.exist()
        problem.addClause([adjs[0],adjs[1],adjv])
        problem.addClause([~adjv,adjs[2], ~nv])
        
        #problem.addClause([~adjs[0],~adjs[1],~adjs[2], ~nv])
        problem.addClause([~adjs[0],~adjs[1], ~nv])
        problem.addClause([~adjs[0],~adjs[2], ~nv])
        problem.addClause([~adjs[1],~adjs[2], ~nv])
      else:
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])

    ## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == ncols-1 ], problem.ttt())

for prod in prods:
  build_pos(prod)


for prod in duals:
  build_neg(prod)

print "p cnf 253 1125"
#start_time2=time.time()
#print >>sys.stderr, start_time2-start_time_prep1, " :time prepare input to SAT for restricted"
problem.to_dimacs()
#end_time2=time.time()
#print >>sys.stderr, end_time2-start_time2, " :SAT elapsed time for restricted"
